//
// Windows向けunzip
//
// unzipが古くてUTF-8に対応してないので、それっぽいパチモンを用意。
//
// 確認環境
// cargo 1.59.0
// Windows 10
//
// 2022.05.28 Takeshi Yoneki
// ostra@asahi.email.ne.jp
// 

use std::env;
use std::fs;
use std::io::prelude::*;
use std::io::BufReader;
use std::io::BufWriter;
use std::io::SeekFrom;
//use std::path::Path;
use std::path::PathBuf;
use zip::read::ZipArchive;	
//use zip::write::ZipWriter;
//use zip::DateTime;	
//use std::io::Bytes;

//use filetime;
//use filetime::FileTime;
//use time::OffsetDateTime;
//use time::UtcOffset;

mod win_char_conv;
mod win_file;

#[derive(Debug)]
struct Config {
    src: String,
    dst: String,
    listup: bool,
}

impl Config {
    fn new() -> Config {
        Config {
            src: String::new(), dst: String::new(), listup: false,
        }
    }

    fn parse_param(&mut self, args: &[String]) {
        if args.len() == 0 {
            return;
        }
        let p = &args[0];
        if p.starts_with("-") {
            match &p[..] {
                "-d" =>{
                    if args.len() > 1 {
                        let v = &args[1];
                        self.dst = v.clone();
                        self.parse_param(&args[2..]);
                    }
                    else {
                        return;
                    }
                },
                "-l" => {
                    self.listup = true;
                    self.parse_param(&args[1..]);
                },
                _ => {
                    self.parse_param(&args[1..]);
                }
            }
        }
        else {
            if self.src.is_empty() {
                self.src = p.clone()
            }
            else if self.dst.is_empty() {
                self.dst = p.clone()
            }
            self.parse_param(&args[1..]);
            return;
        }
    }
}

const UTF8_FLAG: u16 = 0x800;

fn read_zip_general_purpose_bit_flag(path: &str) -> u16 {
    let mut fp = fs::File::open(path).unwrap();
    fp.seek(SeekFrom::Start(6)).unwrap();
    let mut d = [0; 2];
    let n = fp.read(&mut d[..]).unwrap();
    if n == 2 {
        ((d[1] as u16) << 8) | d[0] as u16
    }
    else {
        0
    }
}

fn main() {
    let mut cfg = Config::new();
    let args: Vec<String> = env::args().collect();
    cfg.parse_param(&args[1..]);
    if cfg.src.is_empty() {
        println!("{}", "usage: unzip SRC_ZIP [-d DST_DIR -l]");
        return;
    }
    if cfg.dst.is_empty() {
        cfg.dst = String::from(".");
    }
    //println!("{}", cfg.src);

    let flag = read_zip_general_purpose_bit_flag(&cfg.src);
    let is_utf8 = (flag & UTF8_FLAG) != 0;

    let mut inz = ZipArchive::new(BufReader::new(fs::File::open(&cfg.src).unwrap())).unwrap();
    
    for index in 0..inz.len() {
        let mut entry = inz.by_index(index).unwrap();
        let path = 
            if is_utf8 {
                String::from(entry.enclosed_name().unwrap().to_str().unwrap())
            }
            else {
                let ansi_string:Vec<u8> = Vec::from(entry.name_raw());
                win_char_conv::ansi_to_utf8(&ansi_string)
            };
        println!("{}", path);
        if !cfg.listup {
            let mut file_path = PathBuf::new();
            file_path.push(&cfg.dst);
            file_path.push(&path);
            let dir_path = file_path.parent().unwrap();
            if !dir_path.exists() {
                fs::create_dir_all(&dir_path).unwrap();
            }
            if entry.is_file() {
                if cfg.listup {
                }
                else {
                    let fp = fs::File::create(&file_path).unwrap();
                    let mut bw = BufWriter::new(fp);
                    let mut readed = 0;
                    let mut buffer: [u8; 1024] = [0; 1024];
                    while readed < entry.size() {
                        let rs = entry.read(&mut buffer).unwrap();
                        readed += rs as u64;
                        bw.write(&buffer[0..rs]).unwrap();
                    }
                }
                // Win32APIを直接使う方法。
                let mtime = entry.last_modified();
                win_file::set_file_mtime(file_path.to_str().unwrap(), mtime.year(), mtime.month() as u16, mtime.day() as u16, mtime.hour() as u16, mtime.minute() as u16, mtime.second() as u16);

                // Win32APIを直接使わない方法。
                // let mtime1 = entry.last_modified();
                // let mtime2 = mtime1.to_time().unwrap();
                // let tz = UtcOffset::current_local_offset().unwrap();
                // let mtime3 = mtime2.replace_offset(tz);
                // let mtime4 = mtime3.unix_timestamp();
                // let mtime5 = FileTime::from_unix_time(mtime4, 0);
                // filetime::set_file_mtime(file_path, mtime5).unwrap();
            }
            else if entry.is_dir() {
                fs::create_dir_all(&file_path).unwrap();
            }
        }
    }
    
    //inz.extract(cfg.dst).unwrap();
}
